"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          15-05-2018
//      Created for Project :   FLAME
"""

import pytest
from pyramid import testing
from pyramid.httpexceptions import HTTPBadRequest
from time import sleep
from clmcservice.utilities import CONFIG_ATTRIBUTES, PROCESS_ATTRIBUTE, STATUS_ATTRIBUTE
import os
import signal


class TestAggregator(object):
    """
    A pytest-implementation test for the aggregator API calls
    """

    @pytest.fixture(autouse=True)
    def app_config(self):
        """
        A fixture to implement setUp/tearDown functionality for all tests by initializing configuration structure for the web service
        """

        self.config = testing.setUp()
        self.config.add_settings({'aggregator_running': False, 'aggregator_report_period': 5,
                                  'aggregator_database_name': 'E2EMetrics', 'aggregator_database_url': "http://172.40.231.51:8086"})

        yield

        testing.tearDown()

    def test_GET(self):
        """
        Tests the GET method for the status of the aggregator.
        """

        from clmcservice.views import AggregatorConfig  # nested import so that importing the class view is part of the test itself

        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "Initially aggregator is not running."
        assert self.config.get_settings().get('aggregator_report_period') == 5, "Initial report period is 5 seconds."
        assert self.config.get_settings().get('aggregator_database_name') == 'E2EMetrics', "Initial database name the aggregator uses is E2EMetrics."
        assert self.config.get_settings().get('aggregator_database_url') == "http://172.40.231.51:8086", "Initial aggregator url is http://172.40.231.51:8086"

        request = testing.DummyRequest()
        response = AggregatorConfig(request).get()

        assert response == {'aggregator_report_period': 5,
                            'aggregator_database_name': 'E2EMetrics',
                            'aggregator_database_url': "http://172.40.231.51:8086"}, "Response must be a dictionary representing a JSON object with the correct configuration data of the aggregator."

        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "A GET request must not modify the aggregator configuration data."
        assert self.config.get_settings().get('aggregator_report_period') == 5, "A GET request must not modify the aggregator configuration data."
        assert self.config.get_settings().get('aggregator_database_name') == 'E2EMetrics', "A GET request must not modify the aggregator configuration data."
        assert self.config.get_settings().get('aggregator_database_url') == "http://172.40.231.51:8086", "A GET request must not modify the aggregator configuration data."

    @pytest.mark.parametrize("input_body, output_value", [
        ('{"aggregator_report_period": 10, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://171.40.231.51:8086"}',
         {'aggregator_report_period': 10, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://171.40.231.51:8086"}),
        ('{"aggregator_report_period": 15, "aggregator_database_name": "E2EMetrics", "aggregator_database_url": "http://172.50.231.51:8086"}',
         {'aggregator_report_period': 15, 'aggregator_database_name': "E2EMetrics", 'aggregator_database_url': "http://172.50.231.51:8086"}),
        ('{"aggregator_report_period": 20, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://172.60.231.51:8086"}',
         {'aggregator_report_period': 20, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://172.60.231.51:8086"}),
        ('{"aggregator_report_period": 25, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://172.60.231.51:8086"}',
         {'aggregator_report_period': 25, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://172.60.231.51:8086"}),
        ('{"aggregator_report_period": 200, "aggregator_database_name": "E2EMetrics", "aggregator_database_url": "http://172.50.231.51:8086"}',
         {'aggregator_report_period': 200, 'aggregator_database_name': "E2EMetrics", 'aggregator_database_url': "http://172.50.231.51:8086"}),
        ('{"aggregator_report_period": 150, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://172.40.231.51:8086"}',
         {'aggregator_report_period': 150, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://172.40.231.51:8086"}),
        ("{aggregator_report_period: 2hb5, aggregator_database_name: CLMCMetrics, aggregator_database_url: http://172.60.231.51:8086}", None),
        ("{aggregator_report_period: 250-, aggregator_database_name: CLMCMetrics, aggregator_database_url: http://172.60.231.52:8086}", None),
        ("{}", None),
        ("{aggregator_running: true}", None),
    ])
    def test_PUT(self, input_body, output_value):
        """
        Tests the PUT method for the configuration of the aggregator
        :param input_body: the input form parameter
        :param output_value: the expected output value, None for expecting an Exception
        """

        from clmcservice.views import AggregatorConfig  # nested import so that importing the class view is part of the test itself

        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "Initially aggregator is not running."
        assert self.config.get_settings().get('aggregator_report_period') == 5, "Initial report period is 5 seconds."
        assert self.config.get_settings().get('aggregator_database_name') == 'E2EMetrics', "Initial database name the aggregator uses is E2EMetrics."
        assert self.config.get_settings().get('aggregator_database_url') == "http://172.40.231.51:8086", "Initial aggregator url is http://172.40.231.51:8086"

        request = testing.DummyRequest()

        request.body = input_body.encode(request.charset)

        if output_value is not None:
            response = AggregatorConfig(request).put()
            assert response == output_value, "Response of PUT request must include the new status of the aggregator"

            for attribute in CONFIG_ATTRIBUTES:
                assert self.config.get_settings().get(attribute) == output_value.get(attribute), "Aggregator settings configuration is not updated."
        else:
            error_raised = False
            try:
                AggregatorConfig(request).put()
            except HTTPBadRequest:
                error_raised = True

            assert error_raised, "Error must be raised in case of an invalid argument."

    def test_start(self):
        """
        Tests the case of starting the aggregator through an API call.
        """

        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "Initially aggregator is not running."
        assert self.config.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        request = testing.DummyRequest()
        input_body = '{"action": "start"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {STATUS_ATTRIBUTE: True}, "The aggregator should have been started."
        assert self.config.get_settings().get(STATUS_ATTRIBUTE), "The aggregator should have been started."
        assert self.config.get_settings().get(PROCESS_ATTRIBUTE) is not None, "Aggregator process should have been initialized."

        # kill the started process after the test is over
        pid = request.registry.settings[PROCESS_ATTRIBUTE].pid
        os.kill(pid, signal.SIGTERM)

    def test_stop(self):
        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "Initially aggregator is not running."
        assert self.config.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        # send a start request to trigger the aggregator
        request = testing.DummyRequest()
        input_body = '{"action": "start"}'
        request.body = input_body.encode(request.charset)
        AggregatorController(request).put()
        assert self.config.get_settings().get(PROCESS_ATTRIBUTE) is not None, "Aggregator process should have been initialized."

        # test stopping the aggregator process when it is running
        request = testing.DummyRequest()
        input_body = '{"action": "stop"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {STATUS_ATTRIBUTE: False}, "The aggregator should have been stopped."
        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "The aggregator should have been stopped."
        assert self.config.get_settings().get(PROCESS_ATTRIBUTE) is None, "Aggregator process should have been terminated."

        sleep(2)  # put a 2 seconds timeout so that the aggregator process can terminate

        # test stopping the aggregator process when it is not running
        request = testing.DummyRequest()
        input_body = '{"action": "stop"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {STATUS_ATTRIBUTE: False}, "The aggregator should have been stopped."
        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "The aggregator should have been stopped."
        assert self.config.get_settings().get(PROCESS_ATTRIBUTE) is None, "Aggregator process should have been terminated."

    def test_restart(self):
        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "Initially aggregator is not running."
        assert self.config.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        # test restarting the aggregator process when it is stopped
        request = testing.DummyRequest()
        input_body = '{"action": "restart"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {STATUS_ATTRIBUTE: True}, "The aggregator should have been restarted."
        assert self.config.get_settings().get(STATUS_ATTRIBUTE), "The aggregator should have been restarted."
        assert self.config.get_settings().get('aggregator_process'), "The aggregator process should have been reinitialised."

        # test restarting the aggregator process when it is running
        request = testing.DummyRequest()
        input_body = '{"action": "restart"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {STATUS_ATTRIBUTE: True}, "The aggregator should have been restarted."
        assert self.config.get_settings().get(STATUS_ATTRIBUTE), "The aggregator should have been restarted."
        assert self.config.get_settings().get('aggregator_process'), "The aggregator process should have been reinitialised."

        # kill the started process after the test is over
        pid = request.registry.settings[PROCESS_ATTRIBUTE].pid
        os.kill(pid, signal.SIGTERM)

    @pytest.mark.parametrize("input_body", [
        '{"action": "malformed"}',
        '{"action": true}',
        '{"action": false}',
        '{"action": 1}',
        '{invalid-json}',
        '{"action": "start", "unneeded_argument": false}',
        '{}'
    ])
    def test_malformed_actions(self, input_body):
        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not self.config.get_settings().get(STATUS_ATTRIBUTE), "Initially aggregator is not running."
        assert self.config.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        # test restarting the aggregator process when it is running
        request = testing.DummyRequest()
        input_body = input_body
        request.body = input_body.encode(request.charset)

        error_raised = False
        try:
            AggregatorController(request).put()
        except HTTPBadRequest:
            error_raised = True

        assert error_raised
