"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          15-05-2018
//      Created for Project :   FLAME
"""

from json import loads


CONFIG_ATTRIBUTES = ('aggregator_report_period', 'aggregator_database_name', 'aggregator_database_url')

STATUS_ATTRIBUTE = 'aggregator_running'

PROCESS_ATTRIBUTE = 'aggregator_process'

# # URL regular expression used in the django framework
# url_regex = re.compile(
#     r'^(?:http|ftp)s?://' # http:// or https://
#     r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' # domain...
#     r'localhost|' # localhost...
#     r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|' # ...or ipv4
#     r'\[?[A-F0-9]*:[A-F0-9:]+\]?)' # ...or ipv6
#     r'(?::\d+)?' # optional port
#     r'(?:/?|[/?]\S+)$', re.IGNORECASE)


def validate_config_content(configuration):
    """
    A utility function to validate a configuration string representing a JSON dictionary.

    :param configuration: the configuration string to validate
    :return the validated configuration dictionary object with the values converted to their required type
    :raise AssertionError: if the argument is not a valid configuration
    """

    global CONFIG_ATTRIBUTES

    try:
        configuration = loads(configuration)
    except:
        raise AssertionError("Configuration must be a JSON object.")

    assert len(configuration) == len(CONFIG_ATTRIBUTES), "Configuration mustn't contain more attributes than the required ones."

    for attribute in CONFIG_ATTRIBUTES:
        assert attribute in configuration

    assert type(configuration['aggregator_report_period']) == int, "Report period must be an integer, received {0} instead.".format(configuration.get('aggregator_report_period'))

    assert configuration['aggregator_report_period'] > 0, "Report period must be a positive integer, received {0} instead.".format(configuration.get('aggregator_report_period'))

    return configuration


def validate_action_content(content):
    """
    A utility function to validate a content string representing a JSON dictionary.

    :param content: the content string to validate
    :return: the validated content dictionary
    :raise AssertionError: if the argument is not a valid json content
    """

    try:
        content = loads(content)
    except:
        raise AssertionError("Content must be a JSON object.")

    assert len(content) == 1, "Content mustn't contain more attributes than the required one."

    assert content['action'] in ('start', 'stop', 'restart')

    return content


def generate_e2e_delay_report(path_id, source_sfr, target_sfr, endpoint, sf_instance, delay_forward, delay_reverse, delay_service, time):
    """
    Generates a combined averaged measurement about the e2e delay and its contributing parts.

    :param path_id: The path identifier, which is a bidirectional path ID for the request and the response path
    :param source_sfr: source service router
    :param target_sfr: target service router
    :param endpoint: endpoint of the media component
    :param sf_instance: service function instance (media component)
    :param delay_forward: Path delay (Forward direction)
    :param delay_reverse: Path delay (Reverse direction)
    :param delay_service: the media service component response time
    :param time: measurement timestamp
    :return: a list of dict-formatted reports to post on influx
    """

    result = [{"measurement": "e2e_delays",
               "tags": {
                   "path_ID": path_id,
                   "source_SFR": source_sfr,
                   "target_SFR": target_sfr,
                   "endpoint": endpoint,
                   "sf_instance": sf_instance
               },
               "fields": {
                   "delay_forward": float(delay_forward),
                   "delay_reverse": float(delay_reverse),
                   "delay_service": float(delay_service)
               },
               "time": int(1000000000*time)
               }]

    return result
